import matplotlib.pyplot as plt
import numpy as np


def plot_planar_manipulator(q, p_ACplot):
    """
    Plot manipulator with joint positions q and the end-effector location
    p_ACplot.

    The intended use includes the following:

      - plot_planar_manipulator(q, forward_kinematics(q))
      - plot_planar_manipulator(inverse_kinematics(p), p)

    which helps verify the correctness of the implementations.

    """
    base = [0.0, 0.0]
    p_AB = [np.cos(q[0]), np.sin(q[0])]
    p_AC = [
        np.cos(q[0]) + np.cos(q[0] + q[1]),
        np.sin(q[0]) + np.sin(q[0] + q[1]),
    ]

    plt.figure()

    # Plot first link
    plt.plot([base[0], p_AB[0]], [base[1], p_AB[1]], "k-", linewidth=2)
    # Plot second link
    plt.plot([p_AB[0], p_AC[0]], [p_AB[1], p_AC[1]], "k-", linewidth=2)
    # Plot joint locations
    plt.plot(base[0], base[1], "go")
    plt.plot(p_AB[0], p_AB[1], "bo")
    plt.plot(p_AC[0], p_AC[1], "ro")

    # Plot user provided position
    plt.plot(p_ACplot[0], p_ACplot[1], "kx", markersize=15)

    # Set settings so things visualize nicely
    plt.xlim([-2.2, 2.2])
    plt.ylim([-2.2, 2.2])
    plt.gca().set_aspect("equal", adjustable="box")



import numpy as np
from pydrake.all import PointCloud, Rgba, RigidTransform, RotationMatrix
from scipy.spatial import KDTree


def PoseEstimationGivenCorrespondences(p_Om, p_s, chat):
    """Returns optimal X_O given the correspondences"""
    # Apply correspondences, and transpose data to support numpy broadcasting
    p_Omc = p_Om[:, chat].T
    p_s = p_s.T

    # Calculate the central points
    p_Ombar = p_Omc.mean(axis=0)
    p_sbar = p_s.mean(axis=0)

    # Calculate the "error" terms, and form the data matrix
    merr = p_Omc - p_Ombar
    serr = p_s - p_sbar
    W = np.matmul(serr.T, merr)

    # Compute R
    U, Sigma, Vt = np.linalg.svd(W)
    R = np.matmul(U, Vt)
    if np.linalg.det(R) < 0:
        print("fixing improper rotation")
        Vt[-1, :] *= -1
        R = np.matmul(U, Vt)

    # Compute p
    p = p_sbar - np.matmul(R, p_Ombar)

    return RigidTransform(RotationMatrix(R), p)


def FindClosestPoints(kdtree_A, point_cloud_B):
    """
    Finds the nearest (Euclidean) neighbor in kdtree_A for each
    point in point_cloud_B.
    @param kdtree_A A scipy KDTree containing point cloud A
    @param point_cloud_B A 3xN numpy array of points.
    @return indices An (N, ) numpy array of the indices in point_cloud_A of each
        point_cloud_B point's nearest neighbor.
    """
    indices = np.empty(point_cloud_B.shape[1], dtype=int)

    for i in range(point_cloud_B.shape[1]):
        distance, indices[i] = kdtree_A.query(point_cloud_B[:, i], k=1)

    return indices


def PrintResults(X_O, Xhat_O):
    p = X_O.translation()
    aa = X_O.rotation().ToAngleAxis()
    print(f"True position: {p}")
    print(f"True orientation: {aa}")
    p = Xhat_O.translation()
    aa = Xhat_O.rotation().ToAngleAxis()
    print(f"Estimated position: {p}")
    print(f"Estimated orientation: {aa}")


def IterativeClosestPoint(
    p_Om,
    p_Ws,
    X_Ohat=RigidTransform(),
    X_O=None,
    meshcat=None,
    meshcat_scene_path=None,
    max_iterations=None,
):
    """
    Implements the vanilla ICP algorithm corresponding all scene points to a
    model point.
    @param p_Om A 3xN numpy array of "model" points in the model/object frame.
    @param p_Ws A 3xN numpy array of "scene" points in the world frame.
    @param X_Ohat An RigidTransform() containing the initial guess for X_O.
    @param X_O The true, known, X_O; if provided, then the method will print a
               comparison of the results.
    @param meshcat a Meshcat instance for visualizing the point clouds.
    @param meshcat_scene_path A string under which the point clouds will be
                              published.
    @return X_WOhat The estimated pose of the model in the world frame.
    @return chat The indices of correspondences (and index into the model
                 points) for each scene point.
    """
    Nm = p_Ws.shape[1]
    # Set chat to a value that FindClosestPoints will never return.
    chat_previous = np.zeros(Nm) - 1

    kdtree_Om = KDTree(p_Om.T, copy_data=True)

    if meshcat_scene_path:
        cloud = PointCloud(p_Om.shape[1])
        cloud.mutable_xyzs()[:] = p_Om
        meshcat.SetObject(meshcat_scene_path, cloud, rgba=Rgba(0, 0, 1, 1))

    iterations = 0
    while True:
        chat = FindClosestPoints(kdtree_Om, X_Ohat.inverse() @ p_Ws)
        if np.array_equal(chat, chat_previous):
            # Then I've converged.
            break
        chat_previous = chat
        X_Ohat = PoseEstimationGivenCorrespondences(p_Om, p_Ws, chat)
        if meshcat_scene_path:
            meshcat.SetTransform(meshcat_scene_path, X_Ohat)
        iterations += 1
        if max_iterations and iterations >= max_iterations:
            break

    if X_O:
        PrintResults(X_O, X_Ohat)

    return X_Ohat, chat



import json
import os
import unittest
from runpy import run_module

import nbformat
from gradescope_utils.autograder_utils.json_test_runner import JSONTestRunner
from nbconvert import PythonExporter

grader_throws = False


def set_grader_throws(val):
    global grader_throws
    grader_throws = val


class Grader:
    def __init__(self):
        pass

    @staticmethod
    def grade_from_notebooks(test_cases_list, notebook_ipynb_list, results_json):
        """
        Running notebooks in notebook_ipynb_list and evaluating
        them on test_cases_list. Result is written into results_json
        """
        try:
            notebook_locals_list = []
            for notebook_ipynb in notebook_ipynb_list:
                notebook_locals_list.append(Grader.locals_from_notebook(notebook_ipynb))
        except Exception as e:
            Grader.global_fail_with_error_message(
                "Exception when running file: " + notebook_ipynb + ", " + str(e),
                results_json,
            )
            raise

        # Grade notebook_locals_list on test_cases_list
        Grader.grade_output(test_cases_list, notebook_locals_list, results_json)

    @staticmethod
    def grade_output(test_case_list, notebook_locals_list, results_json):
        """Grading the notebook_locals with the provided test_cases"""
        # setup test suite for gradescope
        suite = unittest.TestSuite()
        for test_case, notebook_locals in zip(test_case_list, notebook_locals_list):
            test_case_names = unittest.defaultTestLoader.getTestCaseNames(test_case)
            for test_case_name in test_case_names:
                suite.addTest(test_case(test_case_name, notebook_locals))

        # run all the tests in the suite
        with open(results_json, "w") as fh:
            JSONTestRunner(stream=fh).run(suite)

    @staticmethod
    def locals_from_notebook(notebook_ipynb):
        """Read, run, return locals of notebook"""
        banned_commands = ["HTML"]

        ipynb = json.load(open(notebook_ipynb))

        for cell in ipynb["cells"]:
            # clearOutput as an output type is not supported
            # by nbformat.reads(). Therefore, filter out
            # any output cells with the clearOutput warning
            if "outputs" in cell.keys():
                if any(
                    "clearOutput" in line["output_type"] for line in cell["outputs"]
                ):
                    cell["outputs"] = []

            # erase test cells, this is optional and useful for debugging
            # to avoid recursions when developing
            if any("## TEST ##" in line for line in cell["source"]):
                cell["source"] = []
            # filter out all the lines with banned commands
            if banned_commands is not None:
                cell["source"] = [
                    line
                    for line in cell["source"]
                    if not any(command in line for command in banned_commands)
                ]

        exporter = PythonExporter()
        source, meta = exporter.from_notebook_node(
            nbformat.reads(json.dumps(ipynb), nbformat.NO_CONVERT)
        )
        with open("./cleaned_notebook.py", "w") as fh:
            fh.write(source)

        notebook_locals = run_module("cleaned_notebook")
        os.system("rm cleaned_notebook.py")
        return notebook_locals

    @staticmethod
    def global_fail_with_error_message(msg, results_json):
        """Error message if no specific"""
        results = {"score": 0.0, "output": msg}

        with open(results_json, "w") as f:
            f.write(
                json.dumps(
                    results,
                    indent=4,
                    sort_keys=True,
                    separators=(",", ": "),
                    ensure_ascii=True,
                )
            )

    @staticmethod
    def print_test_results(results_json):
        """Printing the results.json"""
        # open the json file for reading
        with open(results_json, "r") as fh:
            result = json.load(fh)

        # print total score
        max_score = sum(test["max_score"] for test in result["tests"])
        print("Total score is {}/{}.".format(int(result["score"]), max_score))

        # print partial scores
        for test in result["tests"]:
            print(
                "\nScore for {} is {}/{}.".format(
                    test["name"], int(test["score"]), test["max_score"]
                )
            )

            # print error message if any
            if "output" in test:
                print("- " + test["output"])

        if grader_throws and int(result["score"]) != max_score:
            raise RuntimeError("Grader did not award full points.")



import numpy as np
from pydrake.all import MathematicalProgram, Meshcat, Solve

import manipulation.meshcat_utils as dut

prog = MathematicalProgram()
x = prog.NewContinuousVariables(2)
prog.AddCost(x.dot(x))
prog.AddBoundingBoxConstraint(-2, 2, x)
result = Solve(prog)

meshcat = Meshcat()
X, Y = np.meshgrid(np.linspace(-3, 3, 35), np.linspace(-3, 3, 31))
dut.plot_mathematical_program(meshcat, "test", prog, X, Y, result)



"""
A script for creating a Drake-compatible SDFormat file for a triangle mesh.

Credits: Part of this code has been adopted from Greg Izatt
(https://github.com/gizatt/convex_decomp_to_sdf).
"""

import argparse
import logging
import os
import sys
from pathlib import Path
from typing import List, Tuple, Union

import numpy as np
import trimesh
from lxml import etree as ET


def calc_mesh_com_and_inertia(
    mesh: trimesh.Trimesh,
    mass: float,
    frame: np.ndarray = None,
) -> Tuple[np.ndarray, np.ndarray]:
    """Calculates the mesh's center of mass and moment of inertia by assuming a uniform
    mass density.

    Args:
        mesh (trimesh.Trimesh): The mesh.
        mass (float): The mass in kg of the object.
        frame (np.ndarray, optional): The frame that the moment of inertia should be
        expressed in. If None, the center of mass is used. This is a (4,4) homogenous
        transformation matrix.

    Returns:
        Tuple[np.ndarray, np.ndarray]: A tuple of (center_of_mass, moment_of_inertia):
            center_of_mass: The center of mass of shape (3,).
            moment_of_inertia: The moment of inertia of shape (3,3).
    """
    if not mesh.is_watertight:
        logging.warning(
            "The mesh is not watertight. This might lead to incorrect center of mass "
            + "and moment of inertia computations. However, in practice the values are "
            + "often reasonable regardless."
        )

    volume = mesh.volume
    mesh.density = mass / volume
    moment_of_inertia = (
        mesh.moment_inertia if frame is None else mesh.moment_inertia_frame(frame)
    )
    if frame is None:
        logging.info(f"Calculating the moment of inertia about {mesh.center_mass}")
    return mesh.center_mass, moment_of_inertia


def perform_convex_decomposition(
    mesh: trimesh.Trimesh,
    mesh_name: str,
    mesh_dir: Path,
    preview_with_trimesh: bool,
    **kwargs,
) -> List[Path]:
    """Given a mesh, performs a convex decomposition of it with VHACD. The resulting
    convex parts are saved in a subfolder named `<mesh_filename>_parts`.

    NOTE: The current implementation requires trimesh>=4.0.0. See
    https://github.com/mikedh/trimesh/issues/2045.

    Args:
        mesh (trimesh.Trimesh): The mesh to decompose.
        mesh_name (str): The name of the mesh. It is used for naming the mesh parts
        directory.
        mesh_dir (Path): The path to the directory that the mesh is stored in. This is
        used for creating the mesh parts directory.
        preview_with_trimesh (bool): Whether to open (and block on) a window to preview
        the decomposition.
        **kwargs: The convex decomposition arguments that are passed to VHACD.

    Returns:
        List[Path]: The paths of the convex pieces.
    """
    # Create a subdir for the convex parts
    out_dir = mesh_dir / (mesh_name + "_parts")
    os.makedirs(out_dir, exist_ok=True)

    if preview_with_trimesh:
        logging.info(
            "Showing mesh before convex decomposition. Close the window to proceed."
        )
        mesh.show()

    logging.info(
        "Performing convex decomposition. This might take a couple of minutes for "
        + "complicated meshes. If this runs too long, try decreasing --resolution."
    )
    try:
        convex_pieces = mesh.convex_decomposition(**kwargs)
    except Exception as e:
        logging.error(f"Problem performing decomposition: {e}")
        exit(1)
    if not isinstance(convex_pieces, list):
        convex_pieces = [convex_pieces]

    if preview_with_trimesh:
        # Display the convex decomposition, giving each a random colors
        for part in convex_pieces:
            this_color = trimesh.visual.random_color()
            part.visual.face_colors[:] = this_color
        scene = trimesh.scene.scene.Scene()
        for part in convex_pieces:
            scene.add_geometry(part)

        logging.info(
            f"Showing the mesh convex decomposition into {len(convex_pieces)} parts. "
            + "Close the window to proceed."
        )
        scene.show()

    convex_piece_paths: List[Path] = []
    for i, part in enumerate(convex_pieces):
        piece_name = f"convex_piece_{i:03d}.obj"
        path = out_dir / piece_name
        part.export(path)
        convex_piece_paths.append(path)

    return convex_piece_paths


def create_sdf_from_mesh(
    mesh_path: Path,
    mass: float,
    scale: float,
    is_compliant: bool,
    hydroelastic_modulus: float,
    hunt_crossley_dissipation: Union[float, None],
    mu_dynamic: Union[float, None],
    mu_static: Union[float, None],
    preview_with_trimesh: bool,
    **kwargs,
) -> None:
    """Given a mesh, creates an SDFormat file in the same directory that:
    - Uses the mesh as its visual geometry
    - Performs a convex decomposition of the mesh, and uses those pieces
    as the collision geometry
    - Inserts the center of mass and moment of inertia for the object, calculated from
    the original mesh and a uniform density assumption. The density is computed from
    the mesh volume and the given mass

    Args:
        mesh_path (Path): The path to the mesh.
        mass (float): The mass in kg of the mesh.
        scale (float): Scale factor to convert the specified mesh's coordinates to
        meters.
        is_compliant (bool): Whether the SDFormat file will be used for compliant
        Hydroelastic simulations. The object will behave as rigid Hydroelastic if this
        is not specified.
        hydroelastic_modulus (float): The Hydroelastic Modulus. This is only used if
        `is_compliant` is True. The default value leads to low compliance. See
        https://drake.mit.edu/doxygen_cxx/group__hydroelastic__user__guide.html for how
        to pick a value.
        hunt_crossley_dissipation (Union[float, None]): The optional Hydroelastic
        Hunt-Crossley dissipation (s/m). See
        https://drake.mit.edu/doxygen_cxx/group__hydroelastic__user__guide.html for how
        to pick a value.
        mu_dynamic (Union[float, None]): The coefficient of dynamic friction.
        mu_static (Union[float, None]): The coefficient of static friction.
        preview_with_trimesh (bool): Whether to open (and block on) a window to preview
        the decomposition.
        **kwargs: The VHACD arguments.
    """
    # Construct SDF path
    dir_path = mesh_path.parent
    mesh_name = mesh_path.stem
    sdf_path = dir_path / (mesh_name + ".sdf")

    # Load and scale mesh
    mesh = trimesh.load(mesh_path, skip_materials=True, force="mesh")
    mesh.apply_scale(scale)

    # Generate the SDFormat headers
    root_item = ET.Element("sdf", version="1.7", nsmap={"drake": "drake.mit.edu"})
    model_item = ET.SubElement(root_item, "model", name=mesh_name)
    link_item = ET.SubElement(model_item, "link", name=f"{mesh_name}_body_link")
    pose_item = ET.SubElement(link_item, "pose")
    pose_item.text = "0 0 0 0 0 0"

    # Compute and add the physical properties
    com, inertia = calc_mesh_com_and_inertia(mesh=mesh, mass=mass)
    inertial_item = ET.SubElement(link_item, "inertial")
    mass_item = ET.SubElement(inertial_item, "mass")
    mass_item.text = str(mass)
    com_item = ET.SubElement(inertial_item, "pose")
    com_item.text = f"{com[0]:.5f} {com[1]:.5f} {com[2]:.5f} 0 0 0"
    inertia_item = ET.SubElement(inertial_item, "inertia")
    for i in range(3):
        for j in range(i, 3):
            item = ET.SubElement(inertia_item, "i" + "xyz"[i] + "xyz"[j])
            item.text = f"{inertia[i, j]:.5e}"

    # Add the original mesh as the visual mesh
    visual_mesh_path = mesh_path.relative_to(dir_path)
    visual_item = ET.SubElement(link_item, "visual", name="visual")
    geometry_item = ET.SubElement(visual_item, "geometry")
    mesh_item = ET.SubElement(geometry_item, "mesh")
    uri_item = ET.SubElement(mesh_item, "uri")
    uri_item.text = visual_mesh_path.as_posix()
    scale_item = ET.SubElement(mesh_item, "scale")
    scale_item.text = f"{scale} {scale} {scale}"

    # Compute the VHACD convex decomposition and use it as the collision geometry
    mesh_piece_paths = perform_convex_decomposition(
        mesh=mesh,
        mesh_name=mesh_name,
        mesh_dir=dir_path,
        preview_with_trimesh=preview_with_trimesh,
        **kwargs,
    )
    for i, mesh_piece_path in enumerate(mesh_piece_paths):
        mesh_piece_path = mesh_piece_path.relative_to(dir_path)
        collision_item = ET.SubElement(
            link_item, "collision", name=f"collision_{i:03d}"
        )
        geometry_item = ET.SubElement(collision_item, "geometry")
        mesh_item = ET.SubElement(geometry_item, "mesh")
        uri_item = ET.SubElement(mesh_item, "uri")
        uri_item.text = mesh_piece_path.as_posix()
        ET.SubElement(mesh_item, "{drake.mit.edu}declare_convex")

        # Add proximity properties
        proximity_item = ET.SubElement(
            collision_item, "{drake.mit.edu}proximity_properties"
        )
        if is_compliant:
            ET.SubElement(proximity_item, "{drake.mit.edu}compliant_hydroelastic")
            hydroelastic_moulus_item = ET.SubElement(
                proximity_item, "{drake.mit.edu}hydroelastic_modulus"
            )
            hydroelastic_moulus_item.text = f"{hydroelastic_modulus:.3e}"
        else:
            ET.SubElement(proximity_item, "{drake.mit.edu}rigid_hydroelastic")
        if hunt_crossley_dissipation is not None:
            hunt_crossley_dissipation_item = ET.SubElement(
                proximity_item, "{drake.mit.edu}hunt_crossley_dissipation"
            )
            hunt_crossley_dissipation_item.text = f"{hunt_crossley_dissipation:.3f}"
        if mu_dynamic is not None:
            mu_dynamic_item = ET.SubElement(proximity_item, "{drake.mit.edu}mu_dynamic")
            mu_dynamic_item.text = f"{mu_dynamic:.3f}"
        if mu_static is not None:
            mu_static_item = ET.SubElement(proximity_item, "{drake.mit.edu}mu_static")
            mu_static_item.text = f"{mu_static:.3f}"

    logging.info(f"Writing SDF to {sdf_path}")
    ET.ElementTree(root_item).write(sdf_path, pretty_print=True)


if __name__ == "__main__":
    parser = argparse.ArgumentParser(
        description="Create a Drake-compatible SDFormat file for a triangle mesh."
    )
    parser.add_argument(
        "--mesh",
        type=str,
        required=True,
        help="Path to mesh file.",
    )
    parser.add_argument(
        "--mass",
        type=float,
        required=True,
        help="The mass in kg of the object that is represented by the mesh. This is "
        + "used for computing the moment of inertia.",
    )
    parser.add_argument(
        "--scale",
        type=float,
        default=1.0,
        help="Scale factor to convert the specified mesh's coordinates to meters.",
    )
    parser.add_argument(
        "--compliant",
        action="store_true",
        help="Whether the SDFormat file will be used for compliant Hydroelastic "
        + "simulations. The object will behave as rigid Hydroelastic if this is not "
        + "specified.",
    )
    parser.add_argument(
        "--hydroelastic_modulus",
        type=float,
        default=1.0e8,
        help="The Hydroelastic Modulus. This is only used if --compliant is specified. "
        + "The default value leads to low compliance. See "
        + "https://drake.mit.edu/doxygen_cxx/group__hydroelastic__user__guide.html for "
        + "how to pick a value.",
    )
    parser.add_argument(
        "--hunt_crossley_dissipation",
        type=float,
        default=None,
        help="The Hydroelastic Hunt-Crossley dissipation (s/m). See "
        + "https://drake.mit.edu/doxygen_cxx/group__hydroelastic__user__guide.html for "
        + "how to pick a value.",
    )
    parser.add_argument(
        "--mu_dynamic",
        type=float,
        default=None,
        help="The coefficient of dynamic friction.",
    )
    parser.add_argument(
        "--mu_static",
        type=float,
        default=None,
        help="The coefficient of static friction.",
    )
    parser.add_argument(
        "--resolution",
        type=int,
        default=10000000,
        help="VHACD voxel resolution.",
    )
    parser.add_argument(
        "--maxConvexHulls",
        type=int,
        default=64,
        help="VHACD maximum number of convex hulls/ mesh pieces.",
    )
    parser.add_argument(
        "--minimumVolumePercentErrorAllowed",
        type=float,
        default=1.0,
        help="VHACD minimum allowed volume percentage error.",
    )
    parser.add_argument(
        "--maxRecursionDepth",
        type=int,
        default=10,
        help="VHACD maximum recursion depth.",
    )
    parser.add_argument(
        "--no_shrinkWrap",
        action="store_true",
        help="Whether or not to shrinkwrap the voxel positions to the source mesh on "
        + "output.",
    )
    parser.add_argument(
        "--fillMode",
        type=str,
        default="flood",
        choices=["flood", "raycast", "surface"],
        help="VHACD maximum recursion depth.",
    )
    parser.add_argument(
        "--maxNumVerticesPerCH",
        type=int,
        default=64,
        help="VHACD maximum number of triangles per convex hull.",
    )
    parser.add_argument(
        "--no_asyncACD",
        action="store_true",
        help="Whether or not to run VHACD asynchronously, taking advantage of "
        + "additional cores.",
    )
    parser.add_argument(
        "--minEdgeLength",
        type=int,
        default=2,
        help="VHACD minimum voxel patch edge length.",
    )
    parser.add_argument(
        "--preview",
        action="store_true",
        help="Whether to preview the decomposition.",
    )
    parser.add_argument(
        "--log_level",
        type=str,
        default="INFO",
        choices=["CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG"],
        help="Log level.",
    )

    args = parser.parse_args()
    logging.basicConfig(level=args.log_level)

    mesh_path = Path(os.path.abspath(args.mesh))
    if not mesh_path.exists():
        logging.error(f"No mesh found at {mesh_path}!")
        sys.exit(1)

    mass = args.mass
    if mass <= 0:
        logging.error(f"Got a mass of {mass} kg. Mass must be positive!")
        sys.exit(1)

    is_compliant = args.compliant
    hydroelastic_modulus = args.hydroelastic_modulus
    if is_compliant and hydroelastic_modulus is not None and hydroelastic_modulus < 0:
        logging.error(
            f"Got a Hydroelastic modulus of {hydroelastic_modulus:.3e} Pa. A "
            + "non-negative value is required!"
        )
        sys.exit(1)

    hunt_crossley_dissipation = args.hunt_crossley_dissipation
    if hunt_crossley_dissipation is not None and hunt_crossley_dissipation < 0:
        logging.error(
            f"Got a Hunt-Crossley dissipation of {hunt_crossley_dissipation:.3f} s/m. A "
            + "non-negative value is required!"
        )
        sys.exit(1)

    mu_dynamic = args.mu_dynamic
    if mu_dynamic is not None and mu_dynamic < 0:
        logging.error(
            f"Got a coefficient of dynamic friction of {mu_dynamic:.3f}. A "
            + "non-negative value is required!"
        )
        sys.exit(1)

    mu_static = args.mu_static
    if mu_static is not None and mu_static < 0:
        logging.error(
            f"Got a coefficient of static friction of {mu_static:.3f}. A "
            + "non-negative value is required!"
        )
        sys.exit(1)

    create_sdf_from_mesh(
        mesh_path=mesh_path,
        mass=mass,
        scale=args.scale,
        is_compliant=is_compliant,
        hydroelastic_modulus=hydroelastic_modulus,
        hunt_crossley_dissipation=hunt_crossley_dissipation,
        mu_dynamic=mu_dynamic,
        mu_static=mu_static,
        preview_with_trimesh=args.preview,
        resolution=args.resolution,
        maxConvexHulls=args.maxConvexHulls,
        minimumVolumePercentErrorAllowed=args.minimumVolumePercentErrorAllowed,
        maxRecursionDepth=args.maxRecursionDepth,
        shrinkWrap=not args.no_shrinkWrap,
        fillMode=args.fillMode,
        maxNumVerticesPerCH=args.maxNumVerticesPerCH,
        asyncACD=not args.no_asyncACD,
        minEdgeLength=args.minEdgeLength,
    )



from pydrake.multibody.parsing import Parser
from pydrake.multibody.plant import AddMultibodyPlantSceneGraph
from pydrake.perception import Concatenate
from pydrake.systems.framework import DiagramBuilder

from manipulation.scenarios import AddRgbdSensors
from manipulation.utils import ConfigureParser


def MustardExampleSystem():
    builder = DiagramBuilder()

    # Create the physics engine + scene graph.
    plant, scene_graph = AddMultibodyPlantSceneGraph(builder, time_step=0.0)
    parser = Parser(plant)
    ConfigureParser(parser)
    parser.AddModelsFromUrl("package://manipulation/mustard_w_cameras.dmd.yaml")
    plant.Finalize()

    # Add a visualizer just to help us see the object.
    use_meshcat = False
    if use_meshcat:
        meshcat = builder.AddSystem(MeshcatVisualizer(scene_graph))
        builder.Connect(
            scene_graph.get_query_output_port(),
            meshcat.get_geometry_query_input_port(),
        )

    AddRgbdSensors(builder, plant, scene_graph)

    diagram = builder.Build()
    diagram.set_name("depth_camera_demo_system")
    return diagram


def MustardPointCloud(normals=False, down_sample=True):
    system = MustardExampleSystem()
    context = system.CreateDefaultContext()
    plant = system.GetSubsystemByName("plant")
    plant_context = plant.GetMyMutableContextFromRoot(context)

    pcd = []
    for i in range(3):
        cloud = system.GetOutputPort(f"camera{i}_point_cloud").Eval(context)

        # Crop to region of interest.
        pcd.append(cloud.Crop(lower_xyz=[-0.3, -0.3, -0.3], upper_xyz=[0.3, 0.3, 0.3]))

        if normals:
            pcd[i].EstimateNormals(radius=0.1, num_closest=30)

            camera = plant.GetModelInstanceByName(f"camera{i}")
            body = plant.GetBodyByName("base", camera)
            X_C = plant.EvalBodyPoseInWorld(plant_context, body)
            pcd[i].FlipNormalsTowardPoint(X_C.translation())

    # Merge point clouds.
    merged_pcd = Concatenate(pcd)
    if not down_sample:
        return merged_pcd
    # Down sample.
    down_sampled_pcd = merged_pcd.VoxelizedDownSample(voxel_size=0.005)
    return down_sampled_pcd



import unittest

import numpy as np
import timeout_decorator
from gradescope_utils.autograder_utils.decorators import weight
from pydrake.all import RigidTransform, RotationMatrix


class TestMobileBaseIk(unittest.TestCase):
    def __init__(self, test_name, notebook_locals):
        super().__init__(test_name)
        self.notebook_locals = notebook_locals

    @weight(5)
    @timeout_decorator.timeout(60.0)
    def test_mobile_base_ik(self):
        """Test solve_ik"""
        # yapf: disable
        goal_rotation1 = RotationMatrix([  # noqa
            [1, 0, 0],
            [0, -1, 0],
            [0, 0, -1],
        ])
        # yapf: enable
        goal_position1 = np.array([-0.83, 0.18, 1.4])
        goal_pose1 = RigidTransform(goal_rotation1, goal_position1)

        # yapf: disable
        goal_rotation2 = RotationMatrix([  # noqa
            [1, 0, 0],
            [0, 1, 0],
            [0, 0, 1],
        ])
        # yapf: enable
        goal_position2 = np.array([0.83, 0.18, 1.1])
        goal_pose2 = RigidTransform(goal_rotation2, goal_position2)

        # yapf: disable
        goal_rotation3 = RotationMatrix([  # noqa
            [1, 0, 0],
            [0, 0, -1],
            [0, 1, 0],
        ])
        # yapf: enable
        goal_position3 = np.array([0, 1.13, 1.6])
        goal_pose3 = RigidTransform(goal_rotation3, goal_position3)

        for goal_pose in [goal_pose1, goal_pose2, goal_pose3]:
            q = self.notebook_locals["solve_ik"](goal_pose)

            self.assertTrue(q is not None, "IK failed, no configuration returned!")

            diagram, plant, scene_graph = self.notebook_locals["build_env"]()

            context = diagram.CreateDefaultContext()
            plant_context = plant.GetMyContextFromRoot(context)
            sg_context = scene_graph.GetMyContextFromRoot(context)

            plant.SetPositions(plant_context, q)

            query_object = plant.get_geometry_query_input_port().Eval(plant_context)
            inspector = query_object.inspector()

            pairs = (
                scene_graph.get_query_output_port()
                .Eval(sg_context)
                .inspector()
                .GetCollisionCandidates()
            )
            min_dist = np.inf
            for pair in pairs:
                dist = query_object.ComputeSignedDistancePairClosestPoints(
                    pair[0], pair[1]
                ).distance
                if dist < min_dist:
                    min_dist = dist
                    min_pair_name = (
                        inspector.GetName(inspector.GetFrameId(pair[0])),
                        inspector.GetName(inspector.GetFrameId(pair[1])),
                    )

            self.assertTrue(
                min_dist >= 0.0,
                "Collision present between %s and %s." % min_pair_name,
            )

            gripper_body = plant.GetBodyByName("l_gripper_palm_link")
            ee_pose = gripper_body.EvalPoseInWorld(plant_context)

            self.assertTrue(
                np.linalg.norm(
                    goal_pose.translation() - ee_pose.translation(), ord=np.inf
                )
                <= 0.002,
                "End effector position doesn't match goal position.",
            )
            self.assertTrue(
                goal_pose.rotation()
                .InvertAndCompose(ee_pose.rotation())
                .ToAngleAxis()
                .angle()
                <= 0.02,
                "End effector orientation doesn't match goal orientation.",
            )

        q = self.notebook_locals["solve_ik"](
            goal_pose3,
            max_tries=5,
            fix_base=True,
            base_pose=np.array([-1.23, 0.05, 0]),
        )
        self.assertTrue(
            q is None, "Function solve_ik does not follow fix_base parameter."
        )



import time
import typing
from collections import namedtuple
from functools import partial

import numpy as np
from pydrake.common.value import AbstractValue
from pydrake.geometry import Cylinder, Meshcat, MeshcatVisualizer, Rgba, Sphere
from pydrake.math import RigidTransform, RollPitchYaw, RotationMatrix
from pydrake.multibody.plant import MultibodyPlant
from pydrake.solvers import (
    BoundingBoxConstraint,
    MathematicalProgram,
    MathematicalProgramResult,
)
from pydrake.systems.framework import Context, EventStatus, LeafSystem
from pydrake.trajectories import Trajectory

from manipulation import running_as_notebook


class MeshcatSliders(LeafSystem):
    """
    A system that outputs the values from meshcat sliders.

    An output port is created for each element in the list `slider_names`.
    Each element of `slider_names` must itself be an iterable collection
    (list, tuple, set, ...) of strings, with the names of sliders that have
    *already* been added to Meshcat via Meshcat.AddSlider().

    The same slider may be used in multiple ports.
    """

    def __init__(self, meshcat: Meshcat, slider_names: typing.List[str]):
        LeafSystem.__init__(self)

        self._meshcat = meshcat
        self._sliders = slider_names
        for i, slider_iterable in enumerate(self._sliders):
            port = self.DeclareVectorOutputPort(
                f"slider_group_{i}",
                len(slider_iterable),
                partial(self._DoCalcOutput, port_index=i),
            )
            port.disable_caching_by_default()

    def _DoCalcOutput(self, context, output, port_index):
        for i, slider in enumerate(self._sliders[port_index]):
            output[i] = self._meshcat.GetSliderValue(slider)


# This class is scheduled for removal.  Use the pydrake version of
# MeshcatPoseSliders instead.
class _MeshcatPoseSliders(LeafSystem):
    """
    Provides a set of ipywidget sliders (to be used in a Jupyter notebook) with
    one slider for each of roll, pitch, yaw, x, y, and z.  This can be used,
    for instance, as an interface to teleoperate the end-effector of a robot.

    The optional `pose` input port is used ONLY at initialization; it can be
    used to set the initial pose e.g. from the current pose of a MultibodyPlant
    frame.
    """

    # TODO(russt): Use namedtuple defaults parameter once we are Python >= 3.7.
    Visible = namedtuple("Visible", ("roll", "pitch", "yaw", "x", "y", "z"))
    Visible.__new__.__defaults__ = (True, True, True, True, True, True)
    MinRange = namedtuple("MinRange", ("roll", "pitch", "yaw", "x", "y", "z"))
    MinRange.__new__.__defaults__ = (-np.pi, -np.pi, -np.pi, -1.0, -1.0, -1.0)
    MaxRange = namedtuple("MaxRange", ("roll", "pitch", "yaw", "x", "y", "z"))
    MaxRange.__new__.__defaults__ = (np.pi, np.pi, np.pi, 1.0, 1.0, 1.0)
    Value = namedtuple("Value", ("roll", "pitch", "yaw", "x", "y", "z"))
    Value.__new__.__defaults__ = (0.0, 0.0, 0.0, 0.0, 0.0, 0.0)
    DecrementKey = namedtuple("DecrementKey", ("roll", "pitch", "yaw", "x", "y", "z"))
    DecrementKey.__new__.__defaults__ = (
        "KeyQ",
        "KeyW",
        "KeyA",
        "KeyJ",
        "KeyI",
        "KeyO",
    )
    IncrementKey = namedtuple("IncrementKey", ("roll", "pitch", "yaw", "x", "y", "z"))
    IncrementKey.__new__.__defaults__ = (
        "KeyE",
        "KeyS",
        "KeyD",
        "KeyL",
        "KeyK",
        "KeyU",
    )

    def __init__(
        self,
        meshcat,
        visible=Visible(),
        min_range=MinRange(),
        max_range=MaxRange(),
        value=Value(),
        decrement_keycode=DecrementKey(),
        increment_keycode=IncrementKey(),
        body_index=None,
    ):
        """
        Args:
            meshcat: A Meshcat instance.
            visible: An object with boolean elements for 'roll', 'pitch',
                'yaw', 'x', 'y', 'z'; the intention is for this to be the
                PoseSliders.Visible() namedtuple.  Defaults to all true.
            min_range, max_range, value: Objects with float values for 'roll',
                'pitch', 'yaw', 'x', 'y', 'z'; the intention is for the
                caller to use the PoseSliders.MinRange, MaxRange, and
                Value namedtuples.  See those tuples for default values.
            body_index: if the body_poses input port is connected, then this
                index determine which pose is used to set the initial
                slider positions during the Initialization event.
        """
        LeafSystem.__init__(self)
        port = self.DeclareAbstractOutputPort(
            "pose",
            lambda: AbstractValue.Make(RigidTransform()),
            self._DoCalcOutput,
        )

        self.DeclareAbstractInputPort(
            "body_poses", AbstractValue.Make([RigidTransform()])
        )
        self.DeclareInitializationDiscreteUpdateEvent(self._Initialize)

        # The widgets themselves have undeclared state.  For now, we accept it,
        # and simply disable caching on the output port.
        # TODO(russt): consider implementing the more elaborate methods seen
        # in, e.g., LcmMessageSubscriber.
        port.disable_caching_by_default()

        self._meshcat = meshcat
        self._visible = visible
        self._value = list(value)
        self._body_index = body_index

        print("Keyboard Controls:")
        for i in range(6):
            if visible[i]:
                meshcat.AddSlider(
                    min=min_range[i],
                    max=max_range[i],
                    value=value[i],
                    step=0.02,
                    name=value._fields[i],
                    decrement_keycode=decrement_keycode[i],
                    increment_keycode=increment_keycode[i],
                )
                print(
                    f"{value._fields[i]} : {decrement_keycode[i]} / {increment_keycode[i]}"  # noqa
                )

    def __del__(self):
        for s in ["roll", "pitch", "yaw", "x", "y", "z"]:
            if visible[s]:
                self._meshcat.DeleteSlider(s)

    def SetPose(self, pose):
        """
        Sets the current value of the sliders.

        Args:
            pose: Any viable argument for the RigidTransform
                constructor.
        """
        tf = RigidTransform(pose)
        self.SetRpy(RollPitchYaw(tf.rotation()))
        self.SetXyz(tf.translation())

    def SetRpy(self, rpy):
        """
        Sets the current value of the sliders for roll, pitch, and yaw.

        Args:
            rpy: An instance of drake.math.RollPitchYaw
        """
        self._value[0] = rpy.roll_angle()
        self._value[1] = rpy.pitch_angle()
        self._value[2] = rpy.yaw_angle()
        for i in range(3):
            if self._visible[i]:
                self._meshcat.SetSliderValue(self._visible._fields[i], self._value[i])

    def SetXyz(self, xyz):
        """
        Sets the current value of the sliders for x, y, and z.

        Args:
            xyz: A 3 element iterable object with x, y, z.
        """
        self._value[3:] = xyz
        for i in range(3, 6):
            if self._visible[i]:
                self._meshcat.SetSliderValue(self._visible._fields[i], self._value[i])

    def _update_values(self):
        changed = False
        for i in range(6):
            if self._visible[i]:
                old_value = self._value[i]
                self._value[i] = self._meshcat.GetSliderValue(self._visible._fields[i])
                changed = changed or self._value[i] != old_value
        return changed

    def _get_transform(self):
        return RigidTransform(
            RollPitchYaw(self._value[0], self._value[1], self._value[2]),
            self._value[3:],
        )

    def _DoCalcOutput(self, context, output):
        """Constructs the output values from the sliders."""
        self._update_values()
        output.set_value(self._get_transform())

    def _Initialize(self, context, discrete_state):
        if self.get_input_port().HasValue(context):
            if self._body_index is None:
                raise RuntimeError(
                    "If the `body_poses` input port is connected, then you "
                    "must also pass a `body_index` to the constructor."
                )
            self.SetPose(self.get_input_port().Eval(context)[self._body_index])
            return EventStatus.Succeeded()
        return EventStatus.DidNothing()

    def Run(self, publishing_system, root_context, callback):
        # Calls callback(root_context, pose), then
        # publishing_system.ForcedPublish() each time the sliders change value.
        if not running_as_notebook:
            return

        publishing_context = publishing_system.GetMyContextFromRoot(root_context)

        print("Press the 'Stop PoseSliders' button in Meshcat to continue.")
        self._meshcat.AddButton("Stop PoseSliders", "Escape")
        while self._meshcat.GetButtonClicks("Stop PoseSliders") < 1:
            if self._update_values():
                callback(root_context, self._get_transform())
                publishing_system.ForcedPublish(publishing_context)
            time.sleep(0.1)

        self._meshcat.DeleteButton("Stop PoseSliders")


class WsgButton(LeafSystem):
    """Adds a button named `Open/Close Gripper` to the meshcat GUI, and registers the Space key to press it. Pressing this button will toggle the value of the output port from a wsg position command corresponding to an open position or a closed position.

    Args:
        meshcat: The meshcat instance in which to register the button.
    """

    def __init__(self, meshcat: Meshcat):
        LeafSystem.__init__(self)
        port = self.DeclareVectorOutputPort("wsg_position", 1, self._DoCalcOutput)
        port.disable_caching_by_default()
        self._meshcat = meshcat
        self._button = "Open/Close Gripper"
        meshcat.AddButton(self._button, "Space")
        print("Press Space to open/close the gripper")

    def __del__(self):
        self._meshcat.DeleteButton(self._button)

    def _DoCalcOutput(self, context, output):
        position = 0.107  # open
        if (self._meshcat.GetButtonClicks(self._button) % 2) == 1:
            position = 0.002  # close
        output.SetAtIndex(0, position)


class StopButton(LeafSystem):
    """Adds a button named `Stop Simulation` to the meshcat GUI, and registers
    the `Escape` key to press it. Pressing this button will terminate the
    simulation.

    Args:
        meshcat: The meshcat instance in which to register the button.
        check_interval: The period at which to check for button presses.
    """

    def __init__(self, meshcat: Meshcat, check_interval: float = 0.1):
        LeafSystem.__init__(self)
        self._meshcat = meshcat
        self._button = "Stop Simulation"

        self.DeclareDiscreteState([0])  # button click count
        self.DeclareInitializationDiscreteUpdateEvent(self._Initialize)
        self.DeclarePeriodicDiscreteUpdateEvent(check_interval, 0, self._CheckButton)

        # Create the button now (rather than at initialization) so that the
        # CheckButton method will work even if Initialize has never been
        # called.
        meshcat.AddButton(self._button, "Escape")

    def __del__(self):
        # TODO(russt): Provide a nicer way to check if the button is currently
        # registered.
        try:
            self._meshcat.DeleteButton(self._button)
        except:
            pass

    def _Initialize(self, context, discrete_state):
        print("Press Escape to stop the simulation")
        discrete_state.set_value([self._meshcat.GetButtonClicks(self._button)])

    def _CheckButton(self, context, discrete_state):
        clicks_at_initialization = context.get_discrete_state().value()[0]
        if self._meshcat.GetButtonClicks(self._button) > clicks_at_initialization:
            self._meshcat.DeleteButton(self._button)
            return EventStatus.ReachedTermination(self, "Termination requested by user")
        return EventStatus.DidNothing()


def AddMeshcatTriad(
    meshcat: Meshcat,
    path: str,
    length: float = 0.25,
    radius: float = 0.01,
    opacity: float = 1.0,
    X_PT: RigidTransform = RigidTransform(),
):
    """Adds an X-Y-Z triad to the meshcat scene.

    Args:
        meshcat: A Meshcat instance.
        path: The Meshcat path on which to attach the triad. Using relative paths will attach the triad to the path's coordinate system.
        length: The length of the axes in meters.
        radius: The radius of the axes in meters.
        opacity: The opacity of the axes in [0, 1].
        X_PT: The pose of the triad relative to the path.
    """
    meshcat.SetTransform(path, X_PT)
    # x-axis
    X_TG = RigidTransform(RotationMatrix.MakeYRotation(np.pi / 2), [length / 2.0, 0, 0])
    meshcat.SetTransform(path + "/x-axis", X_TG)
    meshcat.SetObject(
        path + "/x-axis", Cylinder(radius, length), Rgba(1, 0, 0, opacity)
    )

    # y-axis
    X_TG = RigidTransform(RotationMatrix.MakeXRotation(np.pi / 2), [0, length / 2.0, 0])
    meshcat.SetTransform(path + "/y-axis", X_TG)
    meshcat.SetObject(
        path + "/y-axis", Cylinder(radius, length), Rgba(0, 1, 0, opacity)
    )

    # z-axis
    X_TG = RigidTransform([0, 0, length / 2.0])
    meshcat.SetTransform(path + "/z-axis", X_TG)
    meshcat.SetObject(
        path + "/z-axis", Cylinder(radius, length), Rgba(0, 0, 1, opacity)
    )


def plot_mathematical_program(
    meshcat: Meshcat,
    path: str,
    prog: MathematicalProgram,
    X: np.ndarray,
    Y: np.ndarray,
    result: MathematicalProgramResult = None,
    point_size: float = 0.05,
):
    """Visualize a MathematicalProgram in Meshcat.

    Args:
        meshcat: A Meshcat instance.
        path: The Meshcat path to the visualization.
        prog: A MathematicalProgram instance.
        X: A 2D array of x values.
        Y: A 2D array of y values.
        result: A MathematicalProgramResult instance; if provided then the
            solution is visualized as a point.
        point_size: The size of the solution point.
    """
    assert prog.num_vars() == 2
    assert X.size == Y.size

    X.size
    values = np.vstack((X.reshape(-1), Y.reshape(-1)))
    costs = prog.GetAllCosts()

    # Vectorized multiply for the quadratic form.
    # Z = (D*np.matmul(Q,D)).sum(0).reshape(nx, ny)

    if costs:
        Z = prog.EvalBindingVectorized(costs[0], values)
        for b in costs[1:]:
            Z = Z + prog.EvalBindingVectorized(b, values)

    cv = f"{path}/constraints"
    for binding in prog.GetAllConstraints():
        if isinstance(binding.evaluator(), BoundingBoxConstraint):
            c = binding.evaluator()
            var_indices = [
                int(prog.decision_variable_index()[v.get_id()])
                for v in binding.variables()
            ]
            satisfied = np.array(
                c.CheckSatisfiedVectorized(values[var_indices, :], 0.001)
            ).reshape(1, -1)
            if costs:
                Z[~satisfied] = np.nan

            v = f"{cv}/{type(c).__name__}"
            Zc = np.zeros(Z.shape)
            Zc[satisfied] = np.nan
            meshcat.PlotSurface(
                v,
                X,
                Y,
                Zc.reshape(X.shape),
                rgba=Rgba(1.0, 0.2, 0.2, 1.0),
                wireframe=True,
            )
        else:
            Zc = prog.EvalBindingVectorized(binding, values)
            evaluator = binding.evaluator()
            low = evaluator.lower_bound()
            up = evaluator.upper_bound()
            cvb = f"{cv}/{type(evaluator).__name__}"
            for index in range(Zc.shape[0]):
                # TODO(russt): Plot infeasible points in a different color.
                infeasible = np.logical_or(
                    Zc[index, :] < low[index], Zc[index, :] > up[index]
                )
                meshcat.PlotSurface(
                    f"{cvb}/{index}",
                    X,
                    Y,
                    Zc[index, :].reshape(X.shape),
                    rgba=Rgba(1.0, 0.3, 1.0, 1.0),
                    wireframe=True,
                )

    if costs:
        meshcat.PlotSurface(
            f"{path}/objective",
            X,
            Y,
            Z.reshape(X.shape),
            rgba=Rgba(0.3, 1.0, 0.3, 1.0),
            wireframe=True,
        )

    if result:
        v = f"{path}/solution"
        meshcat.SetObject(v, Sphere(point_size), Rgba(0.3, 1.0, 0.3, 1.0))
        x_solution = result.get_x_val()
        meshcat.SetTransform(
            v,
            RigidTransform([x_solution[0], x_solution[1], result.get_optimal_cost()]),
        )


def PublishPositionTrajectory(
    trajectory: Trajectory,
    root_context: Context,
    plant: MultibodyPlant,
    visualizer: MeshcatVisualizer,
    time_step: float = 1.0 / 33.0,
):
    """
    Publishes an animation to Meshcat of a MultibodyPlant using a trajectory of the plant positions.

    Args:
        trajectory: A Trajectory instance.
        root_context: The root context of the diagram containing plant.
        plant: A MultibodyPlant instance.
        visualizer: A MeshcatVisualizer instance.
        time_step: The time step between published frames.
    """
    plant_context = plant.GetMyContextFromRoot(root_context)
    visualizer_context = visualizer.GetMyContextFromRoot(root_context)

    visualizer.StartRecording(False)

    for t in np.append(
        np.arange(trajectory.start_time(), trajectory.end_time(), time_step),
        trajectory.end_time(),
    ):
        root_context.SetTime(t)
        plant.SetPositions(plant_context, trajectory.value(t))
        visualizer.ForcedPublish(visualizer_context)

    visualizer.StopRecording()
    visualizer.PublishRecording()



import gymnasium as gym
import numpy as np
from pydrake.all import (
    AddMultibodyPlantSceneGraph,
    Box,
    ConstantVectorSource,
    ContactVisualizer,
    ContactVisualizerParams,
    DiagramBuilder,
    DiscreteContactApproximation,
    EventStatus,
    FixedOffsetFrame,
    InverseDynamicsController,
    LeafSystem,
    MeshcatVisualizer,
    MultibodyPlant,
    Multiplexer,
    Parser,
    PassThrough,
    PlanarJoint,
    PrismaticJoint,
    RandomGenerator,
    RigidTransform,
    RotationMatrix,
    Simulator,
    SpatialInertia,
    Sphere,
    UnitInertia,
    Variable,
)
from pydrake.gym import DrakeGymEnv

from manipulation.scenarios import AddShape, SetTransparency
from manipulation.utils import ConfigureParser

gym.envs.register(
    id="BoxFlipUp-v0",
    entry_point=("manipulation.envs.box_flipup:BoxFlipUpEnv"),
)

""" Defines the BoxFlipUpEnv

BoxFlipUpEnv is an extremely simple environment, inspired by the example used
in the force control chapter, in which a point finger tries to rotate a box
inside a bin onto its side.
"""


def AddPlanarBinAndSimpleBox(
    plant, mass=1.0, mu=1.0, width=0.2, depth=0.05, height=0.3
):
    parser = Parser(plant)
    ConfigureParser(parser)
    bin = parser.AddModelsFromUrl("package://manipulation/planar_bin.sdf")[0]
    plant.WeldFrames(
        plant.world_frame(),
        plant.GetFrameByName("bin_base", bin),
        RigidTransform(RotationMatrix.MakeZRotation(np.pi / 2.0), [0, 0, -0.015]),
    )

    planar_joint_frame = plant.AddFrame(
        FixedOffsetFrame(
            "planar_joint_frame",
            plant.world_frame(),
            RigidTransform(RotationMatrix.MakeXRotation(np.pi / 2)),
        )
    )

    # TODO(russt): make this a *random* box?
    # TODO(russt): move random box to a shared py file.
    box_instance = AddShape(plant, Box(width, depth, height), "box", mass, mu)
    box_joint = plant.AddJoint(
        PlanarJoint(
            "box_joint",
            planar_joint_frame,
            plant.GetFrameByName("box", box_instance),
        )
    )
    box_joint.set_default_translation([0, height / 2.0])
    return box_instance


def AddPointFinger(plant):
    finger = AddShape(plant, Sphere(0.01), "finger", color=[0.9, 0.5, 0.5, 1.0])
    false_body1 = plant.AddRigidBody(
        "false_body1",
        finger,
        SpatialInertia(0, [0, 0, 0], UnitInertia(0, 0, 0)),
    )
    finger_x = plant.AddJoint(
        PrismaticJoint(
            "finger_x",
            plant.world_frame(),
            plant.GetFrameByName("false_body1"),
            [1, 0, 0],
            -0.3,
            0.3,
        )
    )
    plant.AddJointActuator("finger_x", finger_x)
    finger_z = plant.AddJoint(
        PrismaticJoint(
            "finger_z",
            plant.GetFrameByName("false_body1"),
            plant.GetFrameByName("finger"),
            [0, 0, 1],
            0.0,
            0.5,
        )
    )
    finger_z.set_default_translation(0.25)
    plant.AddJointActuator("finger_z", finger_z)

    return finger


class RewardSystem(LeafSystem):
    def __init__(self):
        LeafSystem.__init__(self)
        self.DeclareVectorInputPort("box_state", 6)
        self.DeclareVectorInputPort("finger_state", 4)
        self.DeclareVectorInputPort("actions", 2)
        self.DeclareVectorOutputPort("reward", 1, self.CalcReward)

    def CalcReward(self, context, output):
        box_state = self.get_input_port(0).Eval(context)
        finger_state = self.get_input_port(1).Eval(context)
        actions = self.get_input_port(2).Eval(context)

        angle_from_vertical = (box_state[2] % np.pi) - np.pi / 2
        cost = 2 * angle_from_vertical**2  # box angle
        cost += 0.1 * box_state[5] ** 2  # box velocity
        effort = actions - finger_state[:2]
        cost += 0.1 * effort.dot(effort)  # effort
        # finger velocity
        cost += 0.1 * finger_state[2:].dot(finger_state[2:])
        # Add 10 to make rewards positive (to avoid rewarding simulator
        # crashes).
        output[0] = 10 - cost


def make_box_flipup(generator, observations="state", meshcat=None, time_limit=10):
    builder = DiagramBuilder()
    plant, scene_graph = AddMultibodyPlantSceneGraph(builder, time_step=0.001)
    plant.set_discrete_contact_approximation(DiscreteContactApproximation.kSap)
    # TODO(russt): randomize parameters.
    box = AddPlanarBinAndSimpleBox(plant)
    finger = AddPointFinger(plant)
    plant.Finalize()
    plant.set_name("plant")
    SetTransparency(scene_graph, alpha=0.5, source_id=plant.get_source_id())
    controller_plant = MultibodyPlant(time_step=0.005)
    AddPointFinger(controller_plant)

    if meshcat:
        MeshcatVisualizer.AddToBuilder(builder, scene_graph, meshcat)
        meshcat.Set2dRenderMode(xmin=-0.35, xmax=0.35, ymin=-0.1, ymax=0.3)
        ContactVisualizer.AddToBuilder(
            builder,
            plant,
            meshcat,
            ContactVisualizerParams(radius=0.005, newtons_per_meter=40.0),
        )

    controller_plant.Finalize()

    # Stiffness control.  (For a point finger with unit mass, the
    # InverseDynamicsController is identical)
    N = controller_plant.num_positions()
    kp = [100] * N
    ki = [1] * N
    kd = [2 * np.sqrt(kp[0])] * N
    controller = builder.AddSystem(
        InverseDynamicsController(controller_plant, kp, ki, kd, False)
    )
    builder.Connect(
        plant.get_state_output_port(finger),
        controller.get_input_port_estimated_state(),
    )

    actions = builder.AddSystem(PassThrough(N))
    positions_to_state = builder.AddSystem(Multiplexer([N, N]))
    builder.Connect(actions.get_output_port(), positions_to_state.get_input_port(0))
    zeros = builder.AddSystem(ConstantVectorSource([0] * N))
    builder.Connect(zeros.get_output_port(), positions_to_state.get_input_port(1))
    builder.Connect(
        positions_to_state.get_output_port(),
        controller.get_input_port_desired_state(),
    )
    builder.Connect(
        controller.get_output_port_control(), plant.get_actuation_input_port()
    )

    builder.ExportInput(actions.get_input_port(), "actions")
    if observations == "state":
        builder.ExportOutput(plant.get_state_output_port(), "observations")
    # TODO(russt): Add 'time', and 'keypoints'
    else:
        raise ValueError("observations must be one of ['state']")

    reward = builder.AddSystem(RewardSystem())
    builder.Connect(plant.get_state_output_port(box), reward.get_input_port(0))
    builder.Connect(plant.get_state_output_port(finger), reward.get_input_port(1))
    builder.Connect(actions.get_output_port(), reward.get_input_port(2))
    builder.ExportOutput(reward.get_output_port(), "reward")

    # Set random state distributions.
    uniform_random = Variable(name="uniform_random", type=Variable.Type.RANDOM_UNIFORM)
    box_joint = plant.GetJointByName("box_joint")
    x, y = box_joint.get_default_translation()
    box_joint.set_random_pose_distribution([0.2 * uniform_random - 0.1 + x, y], 0)

    diagram = builder.Build()
    simulator = Simulator(diagram)

    # Termination conditions:
    def monitor(context):
        if context.get_time() > time_limit:
            return EventStatus.ReachedTermination(diagram, "time limit")
        return EventStatus.Succeeded()

    simulator.set_monitor(monitor)
    return simulator


def BoxFlipUpEnv(observations="state", meshcat=None, time_limit=10):
    simulator = make_box_flipup(
        RandomGenerator(), observations, meshcat=meshcat, time_limit=time_limit
    )
    action_space = gym.spaces.Box(
        low=np.array([-0.5, -0.1]), high=np.array([0.5, 0.6]), dtype=np.float32
    )

    plant = simulator.get_system().GetSubsystemByName("plant")

    # It is unsound to use raw MBP dof limits as observation bounds, because
    # most simulations will violate those limits in practice (in collision, or
    # due to gravity, or in general because no constraint force is incurred
    # except in violation).  However we don't have any other better limits
    # here.  So we broaden the limits by a fixed offset and hope for the best.
    NUM_DOFS = 5
    POSITION_LIMIT_TOLERANCE = np.full((NUM_DOFS,), 0.1)
    VELOCITY_LIMIT_TOLERANCE = np.full((NUM_DOFS,), 0.5)
    if observations == "state":
        low = np.concatenate(
            (
                plant.GetPositionLowerLimits() - POSITION_LIMIT_TOLERANCE,
                plant.GetVelocityLowerLimits() - VELOCITY_LIMIT_TOLERANCE,
            )
        )
        high = np.concatenate(
            (
                plant.GetPositionUpperLimits() + POSITION_LIMIT_TOLERANCE,
                plant.GetVelocityUpperLimits() + VELOCITY_LIMIT_TOLERANCE,
            )
        )
        observation_space = gym.spaces.Box(
            low=np.asarray(low), high=np.asarray(high), dtype=np.float64
        )

    env = DrakeGymEnv(
        simulator=simulator,
        time_step=0.1,
        action_space=action_space,
        observation_space=observation_space,
        reward="reward",
        action_port_id="actions",
        observation_port_id="observations",
    )

    #    from stable_baselines3.common.monitor import Monitor
    #    env = Monitor(env)
    return env



import unittest

import numpy as np
import timeout_decorator
from gradescope_utils.autograder_utils.decorators import weight


class TestPoseEstimation(unittest.TestCase):
    def __init__(self, test_name, notebook_locals):
        super().__init__(test_name)
        self.notebook_locals = notebook_locals

    @weight(4)
    @timeout_decorator.timeout(2.0)
    def test_filtered_pointcloud(self):
        """Test filtered pointcloud"""
        X_WO = self.notebook_locals["X_WO"]
        p_WA = self.notebook_locals["scene_pcl_np_filtered"]
        num_points = p_WA.shape[1]
        p_OA = X_WO.inverse().multiply(p_WA)
        # bounding box from collision geometry:
        # bbox_min = np.array([-0.0375, -0.025, 0])
        # bbox_max = np.array([0.0375, 0.025, 0.05])
        # relax this a bit (the visual geometry is slightly bigger)
        bbox_min = np.array([-0.04, -0.03, -0.01])
        bbox_max = np.array([0.04, 0.03, 0.06])

        in_bbox = np.all(
            (bbox_min[:, None] <= p_OA) & (p_OA <= bbox_max[:, None]), axis=0
        )

        inliers = np.sum(in_bbox)
        outliers = num_points - inliers

        self.assertLessEqual(
            outliers,
            80,
            "You have too many false positives (too many outliers)",
        )
        self.assertGreaterEqual(
            inliers,
            num_points - 170,
            "You have too many false negatives (not enough inliers)",
        )



import unittest

import numpy as np
import timeout_decorator
from gradescope_utils.autograder_utils.decorators import weight


class TestStochasticOptimization(unittest.TestCase):
    def __init__(self, test_name, notebook_locals):
        super().__init__(test_name)
        self.notebook_locals = notebook_locals
        self.gradient_descent = self.notebook_locals["gradient_descent"]

    @weight(2)
    @timeout_decorator.timeout(15.0)
    def test_exact_gradient(self):
        """Test exact gradient function"""
        np.random.seed(7)
        exact_gradient = self.notebook_locals["exact_gradient"]
        f_eval = self.gradient_descent(
            0.1, exact_gradient, initial_x=np.array([2.0, 2.0]), iter=5
        )
        f_eval = np.array(f_eval)

        f_target = np.array(
            [
                [2.000000, 2.000000, 2.466667],
                [1.790000, 1.850000, 1.961246],
                [1.707546, 1.712750, 1.723588],
                [1.653825, 1.584424, 1.539064],
                [1.614487, 1.463857, 1.384269],
                [1.584082, 1.350302, 1.250706],
            ]
        )

        self.assertLessEqual(
            np.linalg.norm(f_target - f_eval),
            1e-2,
            "You have wrong exact gradients.",
        )

    @weight(2)
    @timeout_decorator.timeout(15.0)
    def test_approximated_gradient(self):
        """Test approximated gradient function"""
        np.random.seed(7)
        approximated_gradient = self.notebook_locals["approximated_gradient"]
        f_eval = self.gradient_descent(
            0.1, approximated_gradient, initial_x=np.array([2.0, 2.0]), iter=5
        )
        f_eval = np.array(f_eval)

        f_target = np.array(
            [
                [2.000000, 2.000000, 2.466667],
                [1.279693, 2.198529, 2.209571],
                [1.279187, 2.192247, 2.200421],
                [1.268846, 2.192274, 2.195393],
                [1.268799, 2.100353, 2.067566],
                [1.232492, 2.078929, 2.020906],
            ]
        )

        self.assertLessEqual(
            np.linalg.norm(f_target - f_eval),
            1e-2,
            "You have wrong approximated gradients.",
        )

    @weight(1)
    @timeout_decorator.timeout(15.0)
    def test_approximated_gradient_with_baseline(self):
        """Test approximated gradient with baseline function"""
        np.random.seed(7)
        approximated_gradient_with_baseline = self.notebook_locals[
            "approximated_gradient_with_baseline"
        ]

        def baseline(x):
            return 5

        def reduced_function(x, rate):
            return approximated_gradient_with_baseline(x, rate, baseline)

        f_eval = self.gradient_descent(
            0.1, reduced_function, initial_x=np.array([2.0, 2.0]), iter=5
        )
        f_eval = np.array(f_eval)

        f_target = np.array(
            [
                [2.000000, 2.000000, 2.466667],
                [1.493827, 2.139510, 2.214994],
                [1.497885, 2.189906, 2.289990],
                [1.383463, 2.190205, 2.244283],
                [1.383293, 1.854385, 1.788513],
                [1.509619, 1.928927, 1.927494],
            ]
        )

        self.assertLessEqual(
            np.linalg.norm(f_target - f_eval),
            1e-2,
            "You have wrong approixmated gradients" "with baseline",
        )



# %%
import numpy as np
from rrt_planner.manipulation_station_collision_checker import (
    ManipulationStationCollisionChecker,
)
from rrt_planner.robot import ConfigurationSpace, Range
from rrt_planner.rrt_planning import Problem


class IiwaProblem(Problem):
    def __init__(
        self,
        q_start: np.array,
        q_goal: np.array,
        gripper_setpoint: float,
        left_door_angle: float,
        right_door_angle: float,
        is_visualizing=False,
    ):
        self.gripper_setpoint = gripper_setpoint
        self.left_door_angle = left_door_angle
        self.right_door_angle = right_door_angle
        self.is_visualizing = is_visualizing

        self.collision_checker = ManipulationStationCollisionChecker(
            is_visualizing=is_visualizing
        )

        # Construct configuration space for IIWA.
        plant = self.collision_checker.plant
        nq = 7
        joint_limits = np.zeros((nq, 2))
        for i in range(nq):
            joint = plant.GetJointByName("iiwa_joint_%i" % (i + 1))
            joint_limits[i, 0] = joint.position_lower_limits()
            joint_limits[i, 1] = joint.position_upper_limits()

        range_list = []
        for joint_limit in joint_limits:
            range_list.append(Range(joint_limit[0], joint_limit[1]))

        def l2_distance(q: tuple):
            sum = 0
            for q_i in q:
                sum += q_i**2
            return np.sqrt(sum)

        max_steps = nq * [np.pi / 180 * 2]  # three degrees
        cspace_iiwa = ConfigurationSpace(range_list, l2_distance, max_steps)

        # Call base class constructor.
        Problem.__init__(
            self,
            x=10,  # not used.
            y=10,  # not used.
            robot=None,  # not used.
            obstacles=None,  # not used.
            start=tuple(q_start),
            goal=tuple(q_goal),
            cspace=cspace_iiwa,
        )

    def collide(self, configuration):
        q = np.array(configuration)
        return self.collision_checker.ExistsCollision(
            q,
            self.gripper_setpoint,
            self.left_door_angle,
            self.right_door_angle,
        )

    def run_planner(self, method: str):
        path = None
        if method == "rrt":
            path = self.rrt_planning()
        elif method == "birrt":
            path = self.bidirectional_rrt_planning()
        else:
            raise NotImplementedError

        if path is None:
            print("No path found")
            return None
        else:
            print(
                "Path found with " + str(len(path) - 1) + " movements of distance ",
                self.path_distance(path),
            )
            smooth_path = self.smooth_path(path)
            print(
                "Smoothed path found with "
                + str(len(smooth_path) - 1)
                + " movements of distance ",
                self.path_distance(smooth_path),
            )
            # interpolated smooth path
            spath = []
            for i in range(1, len(smooth_path)):
                spath.extend(self.cspace.path(smooth_path[i - 1], smooth_path[i]))

            # make sure path is collision free
            if any([self.collide(c) for c in spath]):
                print("Collision in smoothed path")
                return None

            return spath

    def visualize_path(self, path):
        # show path in meshcat
        for q in path:
            q = np.array(q)
            self.collision_checker.DrawStation(
                q,
                self.gripper_setpoint,
                self.left_door_angle,
                self.right_door_angle,
            )
            input("next?")



import typing
import unittest
from functools import cache

from pydrake.all import (
    AddFrame,
    AddModel,
    AddWeld,
    IiwaDriver,
    ModelDirective,
    MultibodyPlant,
    RobotDiagram,
    Rotation,
    Transform,
)

from manipulation.directives_tree import DirectivesTree
from manipulation.station import MakeHardwareStation, MakeMultibodyPlant, Scenario


class DirectivesTreeTest(unittest.TestCase):
    @cache
    def get_flattened_directives(self) -> typing.List[ModelDirective]:
        return [
            ModelDirective(
                add_model=AddModel(
                    name="iiwa",
                    file="package://drake_models/iiwa_description/urdf/iiwa14_no_collision.urdf",
                )
            ),
            ModelDirective(
                add_frame=AddFrame(
                    name="iiwa_origin",
                    X_PF=Transform(
                        base_frame="world",
                        translation=[0, 0.765, 0],
                    ),
                )
            ),
            ModelDirective(
                add_weld=AddWeld(
                    parent="iiwa_origin",
                    child="iiwa::base",
                )
            ),
            ModelDirective(
                add_model=AddModel(
                    name="wsg",
                    file="package://drake_models/wsg_50_description/sdf/schunk_wsg_50_with_tip.sdf",
                )
            ),
            ModelDirective(
                add_frame=AddFrame(
                    name="iiwa::wsg_attach",
                    X_PF=Transform(
                        base_frame="iiwa::iiwa_link_7",
                        translation=[0, 0, 0.114],
                        rotation=Rotation.Rpy(deg=[90.0, 0.0, 68.0]),
                    ),
                )
            ),
            ModelDirective(
                add_weld=AddWeld(
                    parent="iiwa::wsg_attach",
                    child="wsg::body",
                )
            ),
            ModelDirective(
                add_model=AddModel(
                    name="table",
                    file="package://drake_models/manipulation_station/table_wide.sdf",
                )
            ),
            ModelDirective(
                add_frame=AddFrame(
                    name="table_origin",
                    X_PF=Transform(
                        base_frame="world",
                        translation=[0.4, 0.3825, 0.0],
                        rotation=Rotation.Rpy(deg=[0.0, 0.0, 0.0]),
                    ),
                )
            ),
            ModelDirective(
                add_weld=AddWeld(
                    parent="table_origin",
                    child="table::table_body",
                )
            ),
        ]

    def test_get_welded_descendants_and_directives(self):
        directives = self.get_flattened_directives()
        tree = DirectivesTree(directives)

        children, wsg_directives = tree.GetWeldedDescendantsAndDirectives(["iiwa"])
        self.assertEqual(children, {"wsg"})
        sorted_wsg_directives = tree.TopologicallySortDirectives(wsg_directives)
        self.assertEqual(
            sorted_wsg_directives, directives[3:6]
        )  # wsg-related directives

    def test_get_weld_to_world_directives(self):
        directives = self.get_flattened_directives()
        tree = DirectivesTree(directives)

        iiwa_directives = tree.GetDirectivesFromRootToModels(["iiwa"])
        sorted_iiwa_directives = tree.TopologicallySortDirectives(iiwa_directives)
        self.assertEqual(
            sorted_iiwa_directives, directives[:3]
        )  # iiwa-related directives

        iiwa_wsg_directives = tree.GetDirectivesFromRootToModels(["iiwa", "wsg"])
        sorted_iiwa_wsg_directives = tree.TopologicallySortDirectives(
            iiwa_wsg_directives
        )
        self.assertEqual(sorted_iiwa_wsg_directives, directives[:6])

    def test_make_multibody_plant(self):
        scenario = Scenario()
        scenario.directives = self.get_flattened_directives()
        scenario.model_drivers = {
            "iiwa": IiwaDriver(
                control_mode="position_only",
                hand_model_name="wsg",
            )
        }

        # Test with all model instances.
        plant: MultibodyPlant = MakeMultibodyPlant(scenario)
        self.assertTrue(plant.HasModelInstanceNamed("iiwa"))
        self.assertTrue(plant.HasModelInstanceNamed("wsg"))
        self.assertTrue(plant.HasModelInstanceNamed("table"))

        # Test with only "iiwa" and "wsg".
        plant: MultibodyPlant = MakeMultibodyPlant(
            scenario, model_instance_names=["iiwa", "wsg"]
        )
        self.assertTrue(plant.HasModelInstanceNamed("iiwa"))
        self.assertTrue(plant.HasModelInstanceNamed("wsg"))
        self.assertFalse(plant.HasModelInstanceNamed("table"))

        # Test with only "iiwa".
        plant: MultibodyPlant = MakeMultibodyPlant(
            scenario, model_instance_names=["iiwa"]
        )
        self.assertTrue(plant.HasModelInstanceNamed("iiwa"))
        self.assertFalse(plant.HasModelInstanceNamed("wsg"))
        self.assertFalse(plant.HasModelInstanceNamed("table"))

        # Test with only "iiwa" and "add_frozen_child_instances=True"
        plant: MultibodyPlant = MakeMultibodyPlant(
            scenario,
            model_instance_names=["iiwa"],
            add_frozen_child_instances=True,
        )
        self.assertTrue(plant.HasModelInstanceNamed("iiwa"))
        self.assertTrue(plant.HasModelInstanceNamed("wsg"))
        self.assertFalse(plant.HasModelInstanceNamed("table"))

    def test_load_scenario(self):
        scenario = Scenario()
        scenario.directives = self.get_flattened_directives()
        scenario.model_drivers = {
            "iiwa": IiwaDriver(
                control_mode="position_only",
                hand_model_name="wsg",
            )
        }
        station: RobotDiagram = MakeHardwareStation(scenario, hardware=False)
        controller_plant: MultibodyPlant = station.GetSubsystemByName(
            "iiwa_controller_plant_pointer_system"
        ).get()

        # Check that the controller plant contains "iiwa" and "wsg" but not "table".
        self.assertTrue(controller_plant.HasModelInstanceNamed("iiwa"))
        self.assertTrue(controller_plant.HasModelInstanceNamed("wsg"))
        self.assertFalse(controller_plant.HasModelInstanceNamed("table"))


if __name__ == "__main__":
    unittest.main()



import unittest

import numpy as np
import timeout_decorator
from gradescope_utils.autograder_utils.decorators import weight


class TestHybrid(unittest.TestCase):
    def __init__(self, test_name, notebook_locals):
        super().__init__(test_name)
        self.notebook_locals = notebook_locals

    @weight(2)
    @timeout_decorator.timeout(30.0)
    def test_antipodal_points(self):
        """Test compute_ctrl"""
        log = self.notebook_locals["log"]
        plant = self.notebook_locals["plant"]
        plant_context = self.notebook_locals["plant_context"]

        # Get final pose of the book.
        X_WB = plant.GetFreeBodyPose(plant_context, plant.GetBodyByName("book"))

        p_WB = X_WB.translation()

        # 1. Check upper bound on book pose.
        self.assertLessEqual(
            p_WB[0],
            0.55,
            "Edge of the book is not between the gap. " "It is too far.",
        )

        # 2. Check lower bound on book pose.
        self.assertGreaterEqual(
            p_WB[0],
            0.35,
            "Edge of the book is not between the gap. " "It is too close.",
        )

        time = log.sample_times()
        traj = log.data()

        finger_length = 0.12

        p_x = traj[1, :] - finger_length * np.sin(traj[0, :])
        p_y = traj[2, :] - finger_length * np.cos(traj[0, :])

        for t in range(len(time)):
            self.assertGreaterEqual(
                p_y[t],
                0.035,
                "The gripper tip cannot be lower than the " "surface of the book.",
            )



"""
This file contains a number of helper systems useful for setting up diagrams.
"""

import numpy as np
from pydrake.all import (
    AbstractValue,
    Body,
    DiagramBuilder,
    DifferentialInverseKinematicsIntegrator,
    DifferentialInverseKinematicsParameters,
    Frame,
    LeafSystem,
    MultibodyPlant,
    RigidTransform,
)


class ExtractPose(LeafSystem):
    """A system that extracts a single pose from a List[RigidTransform], which is the
    output of the MultibodyPlant body_poses output port.

    Args:
        index: The index of the element in the vector whose pose we want to extract
            (e.g. `int(plant.GetBodyByName("body").index())`).
        X_BA: An optional fixed transform from the frame B in the list to an output
            frame A.
    """

    def __init__(
        self,
        index: int,
        X_BA: RigidTransform = RigidTransform(),
    ):
        LeafSystem.__init__(self)
        self.index = index
        self.DeclareAbstractInputPort("poses", AbstractValue.Make([RigidTransform()]))
        self.DeclareAbstractOutputPort(
            "pose",
            lambda: AbstractValue.Make(RigidTransform()),
            self._CalcOutput,
        )
        self.X_BA = X_BA

    def _CalcOutput(self, context, output):
        poses = self.EvalAbstractInput(context, 0).get_value()
        pose = poses[self.index] @ self.X_BA
        output.get_mutable_value().set(pose.rotation(), pose.translation())


class MultibodyPositionToBodyPose(LeafSystem):
    """A system that computes a body pose from a MultibodyPlant position vector. The
    output port calls `plant.SetPositions()` and then `plant.EvalBodyPoseInWorld()`.

    Args:
        plant: The MultibodyPlant.
        body: A body in the plant whose pose we want to compute (e.g. `plant.
            GetBodyByName("body")`).
    """

    def __init__(self, plant: MultibodyPlant, body: Body):
        LeafSystem.__init__(self)
        self.plant = plant
        self.body = body
        self.plant_context = plant.CreateDefaultContext()
        self.DeclareVectorInputPort("position", plant.num_positions())
        self.DeclareAbstractOutputPort(
            "pose",
            lambda: AbstractValue.Make(RigidTransform()),
            self._CalcOutput,
        )

    def _CalcOutput(self, context, output):
        position = self.get_input_port().Eval(context)
        self.plant.SetPositions(self.plant_context, position)
        pose = self.plant.EvalBodyPoseInWorld(self.plant_context, self.body)
        output.get_mutable_value().set(pose.rotation(), pose.translation())


def AddIiwaDifferentialIK(
    builder: DiagramBuilder, plant: MultibodyPlant, frame: Frame = None
) -> DifferentialInverseKinematicsIntegrator:
    """Adds a DifferentialInverseKinematicsIntegrator system to the builder with default parameters suitable for use with the standard 7-link iiwa models or the 3-link planar iiwa models.

    Args:
        builder: The DiagramBuilder to which the system should be added.

        plant: The MultibodyPlant passed to the DifferentialInverseKinematicsIntegrator.

        frame: The frame to use for the end effector command. Defaults to the body
            frame of "iiwa_link_7".

    Returns:
        The DifferentialInverseKinematicsIntegrator system.
    """
    params = DifferentialInverseKinematicsParameters(
        plant.num_positions(), plant.num_velocities()
    )
    time_step = plant.time_step()
    q0 = plant.GetPositions(plant.CreateDefaultContext())
    params.set_nominal_joint_position(q0)
    params.set_end_effector_angular_speed_limit(2)
    params.set_end_effector_translational_velocity_limits([-2, -2, -2], [2, 2, 2])
    if frame is None:
        frame = plant.GetFrameByName("iiwa_link_7")
    if plant.num_positions() == 3:  # planar iiwa
        iiwa14_velocity_limits = np.array([1.4, 1.3, 2.3])
        params.set_joint_velocity_limits(
            (-iiwa14_velocity_limits, iiwa14_velocity_limits)
        )
        # These constants are in body frame
        assert (
            frame.name() == "iiwa_link_7"
        ), "Still need to generalize the remaining planar diff IK params for different frames"  # noqa
        params.set_end_effector_velocity_flag([True, False, False, True, False, True])
    else:
        iiwa14_velocity_limits = np.array([1.4, 1.4, 1.7, 1.3, 2.2, 2.3, 2.3])
        params.set_joint_velocity_limits(
            (-iiwa14_velocity_limits, iiwa14_velocity_limits)
        )
        params.set_joint_centering_gain(10 * np.eye(7))
    differential_ik = builder.AddSystem(
        DifferentialInverseKinematicsIntegrator(
            plant,
            frame,
            time_step,
            params,
            log_only_when_result_state_changes=True,
        )
    )
    return differential_ik



import unittest

import numpy as np
import timeout_decorator
from gradescope_utils.autograder_utils.decorators import weight


class TestSimulationTuning(unittest.TestCase):
    def __init__(self, test_name, notebook_locals):
        super().__init__(test_name)
        self.notebook_locals = notebook_locals

    @timeout_decorator.timeout(2.0)
    @weight(2)
    def test_set_timestep(self):
        """Test test_set_timestep"""
        simulator = self.notebook_locals["simulator_b"]
        diagram = self.notebook_locals["diagram_b"]

        context = simulator.get_context()
        plant = diagram.GetSubsystemByName("plant")
        plant_context = plant.GetMyMutableContextFromRoot(context)
        block_poses = plant.GetPositions(plant_context)
        block1_pos = block_poses[:7][4:]
        block2_pos = block_poses[7:][4:]

        block_pos_range = np.array([[0.0, 1.0], [-0.1, 0.1], [0.03, 0.07]])

        in_range_list1 = []
        in_range_list2 = []
        for i in range(3):
            in_range_1 = block_pos_range[i, 0] < block1_pos[i] < block_pos_range[i, 1]
            in_range_2 = block_pos_range[i, 0] < block2_pos[i] < block_pos_range[i, 1]
            in_range_list1.append(in_range_1)
            in_range_list2.append(in_range_2)

        in_range_all = np.asarray(in_range_list1 + in_range_list2)
        self.assertTrue(
            in_range_all.all(),
            "Final block positions are not in the correct range.",
        )

    @timeout_decorator.timeout(2.0)
    @weight(2)
    def test_set_masses(self):
        """Test test_set_masses"""
        simulator = self.notebook_locals["simulator_c"]
        diagram = self.notebook_locals["diagram_c"]

        context = simulator.get_context()
        plant = diagram.GetSubsystemByName("plant")
        plant_context = plant.GetMyMutableContextFromRoot(context)
        block_poses = plant.GetPositions(plant_context)
        block1_pos = block_poses[:7][4:]
        block2_pos = block_poses[7:][4:]

        block1_pos_range = np.array([[0.07, 0.09], [-0.0004, 0.000], [0.0475, 0.0675]])
        block2_pos_range = np.array([[0.25, 0.30], [-0.0009, -0.0005], [0.0275, 0.05]])

        in_range_list1 = []
        in_range_list2 = []
        for i in range(3):
            in_range_1 = block1_pos_range[i, 0] < block1_pos[i] < block1_pos_range[i, 1]
            in_range_2 = block2_pos_range[i, 0] < block2_pos[i] < block2_pos_range[i, 1]
            in_range_list1.append(in_range_1)
            in_range_list2.append(in_range_2)

        in_range_all = np.asarray(in_range_list1 + in_range_list2)
        self.assertTrue(
            in_range_all.all(),
            "Final block positions are not in the correct range.",
        )

    @timeout_decorator.timeout(2.0)
    @weight(3)
    def test_set_friction(self):
        """Test test_set_friction"""
        simulator = self.notebook_locals["simulator_e"]
        diagram = self.notebook_locals["diagram_e"]

        context = simulator.get_context()
        plant = diagram.GetSubsystemByName("plant")
        plant_context = plant.GetMyMutableContextFromRoot(context)
        block_poses = plant.GetPositions(plant_context)
        block1_pos = block_poses[:7][4:]
        block2_pos = block_poses[7:][4:]

        block1_pos_range = np.array([[0.2, 0.3], [-0.025, 0.025], [0.025, 0.0375]])
        block2_pos_range = np.array([[0.2, 0.3], [-0.025, 0.025], [0.0475, 0.06]])

        in_range_list1 = []
        in_range_list2 = []
        for i in range(3):
            in_range_1 = block1_pos_range[i, 0] < block1_pos[i] < block1_pos_range[i, 1]
            in_range_2 = block2_pos_range[i, 0] < block2_pos[i] < block2_pos_range[i, 1]
            in_range_list1.append(in_range_1)
            in_range_list2.append(in_range_2)

        in_range_all = np.asarray(in_range_list1 + in_range_list2)
        self.assertTrue(
            in_range_all.all(),
            "Final block positions are not in the correct range.",
        )



import unittest

import numpy as np
import timeout_decorator
from gradescope_utils.autograder_utils.decorators import weight


class TestSimpleQP(unittest.TestCase):
    def __init__(self, test_name, notebook_locals):
        super().__init__(test_name)
        self.notebook_locals = notebook_locals

    @weight(4)
    @timeout_decorator.timeout(1.0)
    def test_simple_qp(self):
        """Test Simple QP problem"""
        f = self.notebook_locals["result_submission"]

        # To test, prepare a random matrix so that we don't show the
        # answers directly.
        np.random.seed(7)
        crypt_mat = np.random.rand(8, 3)
        f_eval = crypt_mat.dot(f).squeeze()

        f_target = np.array(
            [
                0.08640122,
                0.20073211,
                0.16338357,
                0.28908734,
                0.15244127,
                0.27599083,
                0.35008611,
                0.31416134,
            ]
        )

        self.assertLessEqual(
            np.linalg.norm(f_target - np.stack(f_eval)),
            1e-6,
            "The answer to the QP is not correct.",
        )



import unittest

import numpy as np
import timeout_decorator
from gradescope_utils.autograder_utils.decorators import weight


class TestDirectJointControl(unittest.TestCase):
    def __init__(self, test_name, notebook_locals):
        super().__init__(test_name)
        self.notebook_locals = notebook_locals

    @weight(5)
    @timeout_decorator.timeout(1.0)
    def test_position(self):
        """Test iiwa_position_measured"""
        f = self.notebook_locals["teleop_2d_direct"]

        q_cmd = np.array([0.1, 0.2, 0.3])
        station, context = f(interactive=False, q_cmd=q_cmd)
        station.ValidateContext(context)

        commanded_pos = station.GetOutputPort("iiwa.position_commanded").Eval(context)

        self.assertLessEqual(
            np.linalg.norm(commanded_pos - q_cmd),
            1e-3,
            "wrong commanded position",
        )



import pydrake.multibody.parsing
import pydrake.multibody.plant
import pydrake.systems.framework

from manipulation.utils import ConfigureParser

filename = "package://manipulation/two_bins_w_cameras.dmd.yaml"

builder = pydrake.systems.framework.DiagramBuilder()
plant, scene_graph = pydrake.multibody.plant.AddMultibodyPlantSceneGraph(
    builder, time_step=0.001
)
parser = pydrake.multibody.parsing.Parser(plant)
ConfigureParser(parser)
parser.AddModelsFromUrl(filename)
plant.Finalize()

diagram = builder.Build()
context = diagram.CreateDefaultContext()



import numpy as np
from pydrake.all import RigidTransform, RotationMatrix


def GraspCandidateCost(
    diagram,
    context,
    cloud,
    wsg_body_index=None,
    plant_system_name="plant",
    scene_graph_system_name="scene_graph",
    adjust_X_G=False,
):
    """
    Args:
        diagram: A diagram containing a MultibodyPlant+SceneGraph that contains
            a free body gripper and any obstacles in the environment that we
            want to check collisions against. It should not include the objects
            in the point cloud; those are handled separately.
        context: The diagram context.  All positions in the context will be
            held fixed *except* the gripper free body pose.
        cloud: a PointCloud in world coordinates which represents candidate
            grasps.
        wsg_body_index: The body index of the gripper in plant.  If None, then
            a body named "body" will be searched for in the plant.

    Returns:
        cost: The grasp cost

    If adjust_X_G is True, then it also updates the gripper pose in the plant
    context.
    """
    plant = diagram.GetSubsystemByName(plant_system_name)
    plant_context = plant.GetMyMutableContextFromRoot(context)
    scene_graph = diagram.GetSubsystemByName(scene_graph_system_name)
    scene_graph_context = scene_graph.GetMyMutableContextFromRoot(context)
    if wsg_body_index:
        wsg = plant.get_body(wsg_body_index)
    else:
        wsg = plant.GetBodyByName("body")
        wsg_body_index = wsg.index()

    X_G = plant.GetFreeBodyPose(plant_context, wsg)

    # Transform cloud into gripper frame
    X_GW = X_G.inverse()
    p_GC = X_GW @ cloud.xyzs()

    # Crop to a region inside of the finger box.
    crop_min = [-0.05, 0.1, -0.00625]
    crop_max = [0.05, 0.1125, 0.00625]
    indices = np.all(
        (
            crop_min[0] <= p_GC[0, :],
            p_GC[0, :] <= crop_max[0],
            crop_min[1] <= p_GC[1, :],
            p_GC[1, :] <= crop_max[1],
            crop_min[2] <= p_GC[2, :],
            p_GC[2, :] <= crop_max[2],
        ),
        axis=0,
    )

    if adjust_X_G and np.sum(indices) > 0:
        p_GC_x = p_GC[0, indices]
        p_Gcenter_x = (p_GC_x.min() + p_GC_x.max()) / 2.0
        X_G.set_translation(X_G @ np.array([p_Gcenter_x, 0, 0]))
        plant.SetFreeBodyPose(plant_context, wsg, X_G)
        X_GW = X_G.inverse()

    query_object = scene_graph.get_query_output_port().Eval(scene_graph_context)

    # Check collisions between the gripper and the sink
    if query_object.HasCollisions():
        cost = np.inf
        return cost

    # Check collisions between the gripper and the point cloud. `margin`` must
    # be smaller than the margin used in the point cloud preprocessing.
    margin = 0.0
    for i in range(cloud.size()):
        distances = query_object.ComputeSignedDistanceToPoint(
            cloud.xyz(i), threshold=margin
        )
        if distances:
            cost = np.inf
            return cost

    n_GC = X_GW.rotation().multiply(cloud.normals()[:, indices])

    # Penalize deviation of the gripper from vertical.
    # weight * -dot([0, 0, -1], R_G * [0, 1, 0]) = weight * R_G[2,1]
    cost = 20.0 * X_G.rotation().matrix()[2, 1]

    # Reward sum |dot product of normals with gripper x|^2
    cost -= np.sum(n_GC[0, :] ** 2)
    return cost


def GenerateAntipodalGraspCandidate(
    diagram,
    context,
    cloud,
    rng,
    wsg_body_index=None,
    plant_system_name="plant",
    scene_graph_system_name="scene_graph",
):
    """
    Picks a random point in the cloud, and aligns the robot finger with the
    normal of that pixel. The rotation around the normal axis is drawn from a
    uniform distribution over [min_roll, max_roll].

    Args:
        diagram: A diagram containing a MultibodyPlant+SceneGraph that contains
            a free body gripper and any obstacles in the environment that we
            want to check collisions against. It should not include the objects
            in the point cloud; those are handled separately.
        context: The diagram context.  All positions in the context will be
            held fixed *except* the gripper free body pose.
        cloud: a PointCloud in world coordinates which represents candidate
            grasps.
        rng: a np.random.default_rng()
        wsg_body_index: The body index of the gripper in plant.  If None,
            then a body named "body" will be searched for in the plant.

    Returns:
        cost: The grasp cost X_G: The grasp candidate
    """
    plant = diagram.GetSubsystemByName(plant_system_name)
    plant_context = plant.GetMyMutableContextFromRoot(context)
    scene_graph = diagram.GetSubsystemByName(scene_graph_system_name)
    scene_graph.GetMyMutableContextFromRoot(context)
    if wsg_body_index:
        wsg = plant.get_body(wsg_body_index)
    else:
        wsg = plant.GetBodyByName("body")
        wsg_body_index = wsg.index()

    if cloud.size() < 1:
        return np.inf, None

    index = rng.integers(0, cloud.size() - 1)

    # Use S for sample point/frame.
    p_WS = cloud.xyz(index)
    n_WS = cloud.normal(index)

    assert np.isclose(
        np.linalg.norm(n_WS), 1.0
    ), f"Normal has magnitude: {np.linalg.norm(n_WS)}"

    Gx = n_WS  # gripper x axis aligns with normal
    # make orthonormal y axis, aligned with world down
    y = np.array([0.0, 0.0, -1.0])
    if np.abs(np.dot(y, Gx)) < 1e-6:
        # normal was pointing straight down.  reject this sample.
        return np.inf, None

    Gy = y - np.dot(y, Gx) * Gx
    Gz = np.cross(Gx, Gy)
    R_WG = RotationMatrix(np.vstack((Gx, Gy, Gz)).T)
    p_GS_G = [0.054 - 0.01, 0.10625, 0]

    # Try orientations from the center out
    min_roll = -np.pi / 3.0
    max_roll = np.pi / 3.0
    alpha = np.array([0.5, 0.65, 0.35, 0.8, 0.2, 1.0, 0.0])
    for theta in min_roll + (max_roll - min_roll) * alpha:
        # Rotate the object in the hand by a random rotation (around the
        # normal).
        R_WG2 = R_WG.multiply(RotationMatrix.MakeXRotation(theta))

        # Use G for gripper frame.
        p_SG_W = -R_WG2.multiply(p_GS_G)
        p_WG = p_WS + p_SG_W

        X_G = RigidTransform(R_WG2, p_WG)
        plant.SetFreeBodyPose(plant_context, wsg, X_G)
        cost = GraspCandidateCost(diagram, context, cloud, adjust_X_G=True)
        X_G = plant.GetFreeBodyPose(plant_context, wsg)
        if np.isfinite(cost):
            return cost, X_G

    return np.inf, None



from datetime import date

from manipulation.mustard_depth_camera_example import MustardExampleSystem
from manipulation.utils import DrakeVersionGreaterThan, SystemHtml

# Test DrakeVersionGreaterThan
DrakeVersionGreaterThan(date(2023, 12, 1))

# Test SystemHtml
system = MustardExampleSystem()
SystemHtml(system)



import dataclasses as dc
import typing

from pydrake.all import ModelDirective, ScopedName


@dc.dataclass(frozen=True)
class Node:
    name: str
    type: typing.Literal["frame", "model"]


@dc.dataclass(frozen=True)
class Edge:
    parent: Node
    child: Node
    directive: ModelDirective  # either AddFrame or AddWeld


class DirectivesTree:
    """
    ===========================================================================
    DirectivesTree: A set of `add_weld` & `add_frame` directives induce a tree
    ===========================================================================

    • Nodes of the tree
      -----------------
        • Each added frame is a node in the tree.
        • Each model instance (identified by its namespace is a node in the tree).

    • Edges of the tree
      -----------------
        • Each `add_weld` directive is a directed edge from its parent
          frame/model to its child frame/model.
        • Each `add_frame` directive is a directed edge from its base
          frame/model to its child frame/model.
    """

    def __init__(self, flattened_directives: typing.List[ModelDirective]):
        self.flattened_directives = flattened_directives

        # Dictionary of all `add_model` directives indexed by the model names.
        self.add_model_directives: typing.Dict[str, ModelDirective] = dict()

        # Names of frames added by `add_frame` directives.
        # The default "world" frame exists implicitly.
        self.frame_names: typing.Set[str] = {"world"}

        # Names of models added by `add_model` directives.
        self.model_names: typing.Set[str] = set()

        # Mapping from parent nodes to the set of its outgoing edges.
        self.edges: typing.Dict[Node, typing.Set[Edge]] = dict()

        # Read node names.
        for d in self.flattened_directives:
            if d.add_model:
                model_name = d.add_model.name
                self.model_names.add(model_name)
                self.add_model_directives[model_name] = d
            if d.add_frame:
                frame_name = d.add_frame.name
                self.frame_names.add(frame_name)

        # Create edges.
        for d in self.flattened_directives:
            if d.add_weld:
                parent_name = d.add_weld.parent
                child_name = d.add_weld.child
                self._AddEdge(parent_name, child_name, d)

            if d.add_frame:
                parent_name = d.add_frame.X_PF.base_frame
                child_name = d.add_frame.name
                self._AddEdge(parent_name, child_name, d)

    def _MakeNode(self, name: str) -> None:
        # Check if this is an added frame.
        if name in self.frame_names:
            return Node(name, "frame")

        # Check if this corresponds to an added model.
        model_name = ScopedName.Parse(name).get_namespace()
        if model_name in self.model_names:
            return Node(model_name, "model")

        raise ValueError(
            f"Node {name} not found in the tree. It neither corresponds to a "
            f"frame {self.frame_names} nor a model instance {self.model_names}."
        )

    def _AddEdge(
        self, parent_name: str, child_name: str, directive: ModelDirective
    ) -> None:
        parent = self._MakeNode(parent_name)
        child = self._MakeNode(child_name)
        edge = Edge(parent, child, directive)
        if parent not in self.edges:
            self.edges[parent] = {edge}
        else:
            self.edges[parent].add(edge)

    def GetWeldedDescendantsAndDirectives(
        self, model_instance_names: typing.List[str]
    ) -> typing.Tuple[typing.Set[str], typing.Set[ModelDirective]]:
        """
        Returns:
            Set[str]: Names of proper descendant models of the input
                `model_instance_names` in this directives tree.
            Set[ModelDirective]: The directives that need to be added to weld
                the proper descendant models to the input model instances.
        """

        def _RecursiveCall(
            node: Node,
        ) -> typing.Tuple[typing.Set[str], typing.Set[ModelDirective]]:
            """
            Args:
                node (Node): The node to start this recursive call from.

            Returns:
                Set[str]: Names of proper descendant models that are welded to
                    the input node.
                Set[ModelDirective]: The directives that need to be added to
                    weld the proper descendant models to the input node.
            """
            descendants: typing.Set[str] = set()
            directives: typing.Set[ModelDirective] = set()

            for edge in self.edges.get(node, set()):
                _descendants, _directives = _RecursiveCall(edge.child)

                # If the child is a model instance, add its AddModel directive,
                # and add its model's name to the set of descendants.
                if edge.child.type == "model":
                    child_model_name = edge.child.name
                    _descendants.add(child_model_name)
                    _directives.add(self.add_model_directives[child_model_name])

                # If the child node has non-zero descendants, add the edge
                # directive that leads to the child node.
                if len(_descendants) > 0:
                    directives.add(edge.directive)

                    # Add the recursion results.
                    descendants.update(_descendants)
                    directives.update(_directives)

            return descendants, directives

        descendants: typing.Set[str] = set()
        directives: typing.Set[ModelDirective] = set()
        for model_instance_name in model_instance_names:
            assert model_instance_name in self.add_model_directives
            node = Node(model_instance_name, "model")
            _descendants, _directives = _RecursiveCall(node)
            descendants.update(_descendants)
            directives.update(_directives)

        # Don't include the input model instances in the descendants set.
        descendants.difference_update(model_instance_names)

        return descendants, directives

    def GetDirectivesFromRootToModels(
        self, model_instance_names: typing.List[str]
    ) -> typing.Set[ModelDirective]:
        """
        Returns:
            Set[ModelDirective]: The directives that need to be added to
                weld all `model_instance_names` to the "world" frame. This is
                the minimal set of directives necessary to support all
                `model_instance_names` in a plant.
        """

        def _RecursiveCall(node: Node) -> typing.Set[ModelDirective]:
            """
            Args:
                node (Node): The node to start this recursive call from.

            Returns:
                Set[ModelDirective]: The directives that need to be added to
                    weld all `model_instance_names` to the "world" frame.
            """
            directives: typing.Set[ModelDirective] = set()

            # Base case: if the node is one of the model instances, add its AddModel
            # directive.
            if node.type == "model" and node.name in model_instance_names:
                directives.add(self.add_model_directives[node.name])

            for edge in self.edges.get(node, set()):
                _directives = _RecursiveCall(edge.child)

                # If the child node has non-zero directives, add the edge
                # directive that leads to the child node.
                if len(_directives) > 0:
                    directives.add(edge.directive)

                    # Add the recursion results.
                    directives.update(_directives)

            return directives

        world_node = Node("world", "frame")
        return _RecursiveCall(world_node)

    def TopologicallySortDirectives(
        self, directives: typing.Set[ModelDirective]
    ) -> typing.List[ModelDirective]:
        """
        This assumes that the flattened directives are in a valid topologically
        sorted order.
        """
        return [d for d in self.flattened_directives if d in directives]



import warnings

from manipulation.meshcat_utils import *  # noqa

warnings.warn(
    "manipulation.meshcat_cpp_utils has been renamed to manipulation.meshcat_utils. This shim will be removed after 2022-12-31.",
    DeprecationWarning,
    stacklevel=2,
)
